# CLI options
import re

CLI_OPT_TRACE = "--trace"

# mandatory parameters

ENV_GCN_MQTT_SERVER_HOST = "GCN_MQTT_SERVER_HOST"
ENV_GCN_MQTT_SERVER_PORT = "GCN_MQTT_SERVER_PORT"
ENV_GCN_MQTT_USER_NAME = "GCN_MQTT_USER_NAME"
ENV_GCN_MQTT_USER_PASSWORD = "GCN_MQTT_USER_PASSWORD"

# optional parameters

ENV_GCN_MQTT_TRANSPORT = "GCN_MQTT_TRANSPORT"
DEFAULT_GCN_MQTT_TRANSPORT = "tcp"

ENV_GCN_MQTT_TLS_MIN_VERSION = "GCN_MQTT_TLS_MIN_VERSION"
DEFAULT_GCN_MQTT_TLS_MIN_VERSION = None  # use python default for version

ENV_GCN_MQTT_TLS_MAX_VERSION = "GCN_MQTT_TLS_MAX_VERSION"
DEFAULT_GCN_MQTT_TLS_MAX_VERSION = None  # use python default for version

ENV_GCN_MQTT_TLS_CIPHERS = "GCN_MQTT_TLS_CIPHERS"
DEFAULT_GCN_MQTT_TLS_CIPHERS = None  # use python default for version

ENV_GCN_MQTT_KEEPALIVE_SECOND = "GCN_MQTT_KEEPALIVE_SECOND"
DEFAULT_GCN_MQTT_KEEPALIVE_SECOND = 60

ENV_GCN_MQTT_CONNECT_TIMEOUT_SECOND = "GCN_MQTT_CONNECT_TIMEOUT_SECOND"
DEFAULT_GCN_MQTT_CONNECT_TIMEOUT_SECOND = 3

ENV_GCN_MQTT_STILL_CONNECTING_ALERT = "GCN_MQTT_STILL_CONNECTING_ALERT"
DEFAULT_GCN_MQTT_STILL_CONNECTING_ALERT = 5 * 60

ENV_GCN_MQTT_RECONNECT = "GCN_MQTT_RECONNECT"
DEFAULT_GCN_MQTT_RECONNECT = False

ENV_GCN_MQTT_LOG_LEVEL = "GCN_MQTT_LOG_LEVEL"
DEFAULT_GCN_MQTT_LOG_LEVEL = "warning"

ENV_GCN_MQTT_CLIENT_ID_RANDOM_BYTES = "GCN_MQTT_CLIENT_ID_RANDOM_BYTES"
DEFAULT_GCN_MQTT_CLIENT_ID_RANDOM_BYTES = 4

ENV_GCN_MQTT_SOCKET_SEND_BUFFER_SIZE = "GCN_MQTT_SOCKET_SEND_BUFFER_SIZE"
DEFAULT_GCN_MQTT_SOCKET_SEND_BUFFER_SIZE = 2**11

# nothing fast is required, used for the queuing mqtt keepalives and message tasks cleanups
ENV_GCN_IDLE_LOOP_SLEEP_SEC = "GCN_IDLE_LOOP_SLEEP_SEC"
DEFAULT_GCN_IDLE_LOOP_SLEEP_SEC = 1

ENV_GCN_HEARTBEAT_MAX_SKEW_SEC = "GCN_HEARTBEAT_MAX_SKEW_SEC"
DEFAULT_GCN_HEARTBEAT_MAX_SKEW_SEC = 10 * 60

ENV_GCN_HEARTBEAT_WATCHDOG_SEC = "GCN_HEARTBEAT_WATCHDOG_SEC"
DEFAULT_GCN_HEARTBEAT_WATCHDOG_SEC = 60

ENV_GCN_CLIENT_GPIO_CHANGE_DEBOUNCE_SEC = "GCN_CLIENT_GPIO_CHANGE_DEBOUNCE_SEC"
DEFAULT_GCN_CLIENT_GPIO_CHANGE_DEBOUNCE_SEC = 0

# notifications

ENV_GCN_MANAGER_STARTING_RECIPIENTS = "GCN_MANAGER_STARTING_RECIPIENTS"
ENV_GCN_MANAGER_STILL_CONNECTING_RECIPIENTS = "GCN_MANAGER_STILL_CONNECTING_RECIPIENTS"
ENV_GCN_MANAGER_CONNECTED_RECIPIENTS = "GCN_MANAGER_CONNECTED_RECIPIENTS"
ENV_GCN_MANAGER_DISCONNECTED_RECIPIENTS = "GCN_MANAGER_DISCONNECTED_RECIPIENTS"
ENV_GCN_MANAGER_EXITING_RECIPIENTS = "GCN_MANAGER_EXITING_RECIPIENTS"

ENV_GCN_CLIENT_SKEWED_HEARTBEAT_RECIPIENTS = "GCN_CLIENT_HEARTBEAT_SKEWED_RECIPIENTS"
ENV_GCN_CLIENT_MISSED_HEARTBEAT_RECIPIENTS = "GCN_CLIENT_MISSED_HEARTBEAT_RECIPIENTS"
ENV_GCN_CLIENT_DROPPED_ITEMS_RECIPIENTS = "GCN_CLIENT_DROPPED_ITEMS_RECIPIENTS"
ENV_GCN_CLIENT_STATUS_CHANGE_ONLINE_RECIPIENTS = (
    "GCN_CLIENT_STATUS_CHANGE_ONLINE_RECIPIENTS"
)
ENV_GCN_CLIENT_STATUS_CHANGE_OFFLINE_RECIPIENTS = (
    "GCN_CLIENT_STATUS_CHANGE_OFFLINE_RECIPIENTS"
)
ENV_GCN_CLIENT_GPIO_CHANGE_UP_RECIPIENTS = "GCN_CLIENT_GPIO_CHANGE_UP_RECIPIENTS"
ENV_GCN_CLIENT_GPIO_CHANGE_DOWN_RECIPIENTS = "GCN_CLIENT_GPIO_CHANGE_DOWN_RECIPIENTS"

# providers

ENV_GCN_ENABLE_NOTIFY_SMS = "GCN_ENABLE_NOTIFY_SMS"
ENV_GCN_ENABLE_NOTIFY_EMAIL = "GCN_ENABLE_NOTIFY_EMAIL"
ENV_GCN_ENABLE_NOTIFY_TWITTER = "GCN_ENABLE_NOTIFY_TWITTER"

ENV_GCN_EMAIL_FROM = "GCN_EMAIL_FROM"
ENV_GCN_EMAIL_SMTP_HOST = "GCN_EMAIL_SMTP_HOST"
ENV_GCN_EMAIL_SMTP_PORT = "GCN_EMAIL_SMTP_PORT"
ENV_GCN_EMAIL_SMTP_USERNAME = "GCN_EMAIL_SMTP_USERNAME"
ENV_GCN_EMAIL_SMTP_PASSWORD = "GCN_EMAIL_SMTP_PASSWORD"
ENV_GCN_EMAIL_SMTP_STARTTLS = "GCN_EMAIL_SMTP_STARTTLS"
ENV_GCN_EMAIL_SMTP_DEBUG = "GCN_EMAIL_SMTP_DEBUG"
ENV_GCN_EMAIL_SMTP_DISABLE_TLS13 = "GCN_EMAIL_SMTP_DISABLE_TLS13"
ENV_GCN_EMAIL_SMTP_SEND_AS_REPORT = "GCN_EMAIL_SMTP_SEND_AS_REPORT"
ENV_GCN_EMAIL_SMTP_TIMEOUT_SEC = "GCN_EMAIL_SMTP_TIMEOUT_SEC"
DEFAULT_GCN_EMAIL_SMTP_TIMEOUT_SEC = 3

ENV_GCN_ALLOW_SMS_COUNTRIES = "GCN_ALLOW_SMS_COUNTRIES"
ENV_GCN_OVH_SMS_SENDER_NAME = "GCN_OVH_SMS_SENDER_NAME"
ENV_GCN_OVH_SMS_SERVICE_NAME = "GCN_OVH_SMS_SERVICE_NAME"
ENV_GCN_OVH_SMS_USER_NAME = "GCN_OVH_SMS_USER_NAME"
ENV_OVH_API_ENDPOINT_NAME = "OVH_ENDPOINT"
ENV_OVH_API_APPLICATION_KEY = "OVH_APPLICATION_KEY"
ENV_OVH_API_APPLICATION_SECRET = "OVH_APPLICATION_SECRET"
ENV_OVH_API_CONSUMER_KEY = "OVH_CONSUMER_KEY"
ENV_GCN_OVH_API_TIMEOUT_SEC = "GCN_OVH_API_TIMEOUT_SEC"
DEFAULT_GCN_OVH_API_TIMEOUT_SEC = 3

OVH_API_MAX_TIME_SKEW = 3
# https://github.com/ovh/python-ovh/blob/master/ovh/client.py#L67
KNOWN_OVH_API_ENDPOINT_NAMES = (
    "ovh-eu",
    "ovh-us",
    "ovh-ca",
    "kimsufi-eu",
    "kimsufi-ca",
    "soyoustart-eu",
    "soyoustart-ca",
)

ENV_GCN_TWITTER_APP_CONSUMER_KEY = "GCN_TWITTER_APP_CONSUMER_KEY"
ENV_GCN_TWITTER_APP_CONSUMER_SECRET = "GCN_TWITTER_APP_CONSUMER_SECRET"
ENV_GCN_TWITTER_USER_ACCESS_TOKEN = "GCN_TWITTER_USER_ACCESS_TOKEN"
ENV_GCN_TWITTER_USER_ACCESS_TOKEN_SECRET = "GCN_TWITTER_USER_ACCESS_TOKEN_SECRET"

# hardcoded values from Arduino program

MQTT_APP_CLIENT = "gcn"
MQTT_APP_CLIENT_ID = re.compile(r"^(?:[a-zA-Z0-9]{2}:){5}[a-zA-Z0-9]{2}$")
MQTT_APP_CLIENT_OUT = "out"
MQTT_APP_CLIENT_IN = "in"
MQTT_APP_CLIENT_STATUS = "status"
MQTT_APP_CLIENT_GPIO = "gpio"
MQTT_APP_CLIENT_MONITORED_GPIO = "monitored_gpio"
MQTT_APP_CLIENT_HEARTBEAT = "heartbeat"
MQTT_APP_CLIENT_BUFFER_TOTAL_DROPPED_ITEM = "buffer_total_dropped_item"
MQTT_APP_CLIENT_SUBSCRIPTION = f"{MQTT_APP_CLIENT}/+/{MQTT_APP_CLIENT_OUT}"

# hardcoded values from Brain program

MQTT_APP_MANAGER = "gcn_manager"
MQTT_APP_MANAGER_ID = re.compile(r"^gcn_manager_[a-zA-Z0-9]+$")
MQTT_APP_MANAGER_STATUS = "status"
MQTT_APP_MANAGER_STATUS_ONLINE = "online"
MQTT_APP_MANAGER_STATUS_OFFLINE = "offline"
MQTT_APP_MANAGER_STATUS_TOPIC = f"{MQTT_APP_MANAGER}/{MQTT_APP_MANAGER_STATUS}"
MQTT_APP_MANAGER_STATUS_SUBSCRIPTION = f"{MQTT_APP_MANAGER_STATUS_TOPIC}/+"



import asyncio
import logging
import time
from argparse import Namespace
from collections import deque
from datetime import timedelta

from pydantic import BaseModel, ValidationError

from gcn_manager import (
    ClientInfo,
    ClientStatus,
    DelayedAction,
    MessageProcessor,
    MqttPublisher,
    MqttMessage,
    MessageError,
    datetime_tz,
)
from gcn_manager.constants import *
from gcn_manager.notifiers import (
    GcnHeartbeatSkewed,
    GcnStatusChangeOnline,
    GcnStatusChangeOffline,
    GcnDroppedItems,
    GcnHeartbeatMissed,
    GcnGpioChangeUp,
    GcnGpioChangeDown,
)


class GpioSample(BaseModel, extra="forbid"):
    state: int
    timestamp: int


class Brain(MessageProcessor):
    def __init__(self, args: Namespace) -> None:
        self._args = args
        self._client_infos: dict[str, ClientInfo] = dict()

    def _ensure_client(self, client_id: str) -> ClientInfo:
        try:
            return self._client_infos[client_id]
        except KeyError:
            pass
        logging.info(f"First time seeing client {client_id}")
        client = ClientInfo(id=client_id)
        notification = GcnHeartbeatMissed(
            client=client, elapsed_seconds=self._args.client_heartbeat_watchdog
        )
        client.heartbeat_watchdog = DelayedAction(
            self._args.client_heartbeat_watchdog,
            notification.send,
            loop=asyncio.get_running_loop(),
            name=f"heartbeat_watchdog_{client_id}",
        )
        self._client_infos[client_id] = client
        return client

    async def process(self, message: MqttMessage, *, publisher: MqttPublisher) -> None:
        logging.debug(f"Processing {message.topic} with payload {message.payload}")
        parts = deque(message.topic.split("/"))
        try:
            app = parts.popleft()

            if app == MQTT_APP_CLIENT:
                id_ = parts.popleft()
                if MQTT_APP_CLIENT_ID.fullmatch(id_) is None:
                    raise MessageError(f"Invalid client id {id_}")
                client = self._ensure_client(id_)
                direction = parts.popleft()

                if direction == MQTT_APP_CLIENT_OUT:
                    category = parts.popleft()

                    if category == MQTT_APP_CLIENT_STATUS:
                        status = message.decode_payload_as(ClientStatus)
                        if client.status is not None and client.status == status:
                            return
                        logging.info(
                            f"Client {client.id} {category} changed : "
                            f"{client.status} -> {status}"
                        )
                        client.status = status
                        if client.status == ClientStatus.OFFLINE:
                            await GcnStatusChangeOffline(client=client).send()
                        elif client.status == ClientStatus.ONLINE:
                            await GcnStatusChangeOnline(client=client).send()
                        else:
                            raise NotImplementedError()  # future guard

                    elif category == MQTT_APP_CLIENT_HEARTBEAT:
                        heartbeat = message.decode_payload_as(int)
                        logging.debug(f"Got heartbeat {heartbeat} for {client.id}")
                        client.heartbeat = heartbeat
                        await client.heartbeat_watchdog.restart()
                        skew = abs(heartbeat - time.time())
                        if skew < self._args.client_heartbeat_max_skew:
                            return
                        await GcnHeartbeatSkewed(
                            client=client,
                            skew=skew,
                            max_skew=self._args.client_heartbeat_max_skew,
                        ).send()

                    elif category == MQTT_APP_CLIENT_BUFFER_TOTAL_DROPPED_ITEM:
                        dropped = message.decode_payload_as(int)
                        if (
                            client.buffer_total_dropped_item is not None
                            and client.buffer_total_dropped_item == dropped
                        ):
                            return
                        logging.info(
                            f"Client {client.id} {category} changed : "
                            f"{client.buffer_total_dropped_item} -> {dropped}"
                        )
                        client.buffer_total_dropped_item = dropped
                        await GcnDroppedItems(client=client).send()

                    elif category == MQTT_APP_CLIENT_MONITORED_GPIO:
                        monitored_gpio = message.decode_payload()
                        monitored_gpio = tuple(
                            sorted(
                                set(pin.strip() for pin in monitored_gpio.split(","))
                            )
                        )
                        if monitored_gpio != client.monitored_gpio:
                            logging.info(
                                f"Client {client.id} monitored GPIO has changed : "
                                f"{client.monitored_gpio}-> {monitored_gpio}"
                            )
                        client.monitored_gpio = monitored_gpio
                        # notify non-monitored gpio for which data has been seen
                        if client.gpio is None:
                            return
                        unmonitored_gpio = tuple(
                            gpio for gpio in client.gpio if gpio not in monitored_gpio
                        )
                        if len(unmonitored_gpio) == 0:
                            return
                        logging.warning(
                            f"Client {client.id} monitors gpio {client.monitored_gpio} "
                            f"but also publishes for gpio {unmonitored_gpio}"
                        )

                    elif category == MQTT_APP_CLIENT_GPIO:
                        gpio = parts.popleft()
                        # notify non-monitored gpio for which data is seen
                        if (
                            client.monitored_gpio is not None
                            and gpio not in client.monitored_gpio
                        ):
                            logging.warning(
                                f"Ignoring gpio {gpio} for client {client.id} which does not monitor it"
                            )
                            return
                        # validate payload
                        sample = message.decode_payload().split()
                        try:
                            if len(sample) != 2:
                                raise ValidationError("Sample must have 2 elements")
                            sample = GpioSample(
                                **{"state": sample[0], "timestamp": sample[1]}
                            )
                        except ValidationError:
                            raise MessageError(
                                f"Client {client.id} gpio {gpio} : invalid payload {message.payload}"
                            )
                        # check if previous
                        try:
                            previous = client.gpio[gpio]
                        except KeyError:
                            when = datetime_tz(sample.timestamp)
                            age = round(abs(sample.timestamp - time.time()))
                            logging.info(
                                f"Client {client.id} gpio {gpio} first sample has state {sample.state} "
                                f"dating from {when} with age of {age} seconds ({timedelta(seconds=age)} ago)"
                            )
                            client.gpio[gpio] = sample
                            return
                        client.gpio[gpio] = sample
                        # check if changed
                        if previous.state == sample.state:
                            logging.debug(
                                f"Client {client.id} gpio {gpio} has not changed from timestamps "
                                f"{previous.timestamp} to {sample.timestamp} : {sample.state}"
                            )
                            return
                        # check if debounce is needed
                        if abs(sample.timestamp - previous.timestamp) < self._args.client_gpio_change_debounce_sec:
                            logging.warning(
                                f"Client {client.id} gpio {gpio} has changed at {sample.timestamp} from previous {previous.timestamp} "
                                f"which is faster than {self._args.client_gpio_change_debounce_sec} sec : skipping notification"
                            )
                            return
                        # notify of change
                        if bool(sample.state):
                            await GcnGpioChangeUp(client=client, gpio_name=gpio).send()
                        else:
                            await GcnGpioChangeDown(
                                client=client, gpio_name=gpio
                            ).send()

                    else:
                        logging.debug(
                            f"Ignoring unknown client topic category {category}"
                        )

                elif direction == MQTT_APP_CLIENT_IN:
                    raise NotImplementedError()

                else:
                    raise MessageError(f"Unknown direction {direction}")

            elif app == MQTT_APP_MANAGER:
                category = parts.popleft()

                if category == MQTT_APP_MANAGER_STATUS:
                    id_ = parts.popleft()
                    if MQTT_APP_MANAGER_ID.fullmatch(id_) is None:
                        raise MessageError(f"Invalid manager id {id_}")
                    if len(message.payload) == 0:
                        logging.debug(
                            f"Empty manager status message for {id_}, ignoring cleanup"
                        )
                        return
                    status = message.decode_payload()
                    if status == MQTT_APP_MANAGER_STATUS_ONLINE:
                        logging.debug(f"Manager {id_} detected online")
                        return
                    elif status == MQTT_APP_MANAGER_STATUS_OFFLINE:
                        logging.debug(
                            f"Manager {id_} detected offline, clearing its status"
                        )
                        publisher.clear_topic(message.topic, qos=1)
                    else:
                        raise MessageError(f"Unknown manager status {status}")

                else:
                    raise MessageError(f"Unknown manager category {category}")

            else:
                raise MessageError(f"Unknown app {app}")

        except IndexError:
            raise MessageError(f"Malformed topic {message.topic}")

    async def shutdown(self) -> None:
        logging.info(
            f"Brain shutting down {len(self._client_infos)} known GCN client(s)"
        )
        while self._client_infos:
            client_id, client_info = self._client_infos.popitem()
            if client_info.heartbeat_watchdog is None:
                continue
            await client_info.heartbeat_watchdog.cancel()
        logging.info(f"Brain shutdown finished")



import asyncio
import enum
import inspect
import logging
import os
from datetime import datetime
from typing import Callable, Any

from pydantic import BaseModel, Field


class NotificationError(Exception):
    pass


class MessageError(Exception):
    pass


class AppError(Exception):
    pass


class MqttMessage(BaseModel):
    topic: str
    payload: bytes

    def decode_payload(self) -> str:
        try:
            return self.payload.decode("utf-8")
        except UnicodeDecodeError:
            raise MessageError(
                f"Could not decode payload {self.payload} for topic {self.topic}"
            )

    def decode_payload_as(self, target_type: type) -> Any:
        decode = self.decode_payload()
        try:
            return target_type(decode)
        except ValueError as e:
            raise MessageError(
                f"Invalid payload {self.payload} for type "
                f"{target_type.__name__} for topic {self.topic} : {e}"
            )


class ClientStatus(enum.StrEnum):
    ONLINE = "online"
    OFFLINE = "offline"


def get_env(key: str) -> str:
    try:
        return os.environ[key]
    except KeyError:
        raise AppError(f"Environment variable {key} not set")


def datetime_system_tz() -> datetime:
    return datetime.now().astimezone()


def datetime_tz(timestamp: float) -> datetime:
    return datetime.fromtimestamp(timestamp).astimezone()


class MqttPublisher:

    def __init__(self, client_id: str) -> None:
        self._client_id = client_id

    @property
    def client_id(self) -> str:
        return self._client_id

    def subscribe(self, topic: str, qos: int) -> None:
        raise NotImplementedError()

    def unsubscribe(self, topic: str) -> None:
        raise NotImplementedError()

    def publish(
        self, topic: str, payload: bytes | bytearray, qos: int = 0, retain: bool = False
    ) -> None:
        raise NotImplementedError()

    def clear_topic(self, topic: str, qos: int = 0) -> None:
        # need retain to remove a retained message
        self.publish(topic, b"", qos=qos, retain=True)


class MessageProcessor:

    async def process(self, message: MqttMessage, *, publisher: MqttPublisher) -> None:
        raise NotImplementedError()

    async def shutdown(self) -> None:
        raise NotImplementedError()


class DelayedAction:
    def __init__(
        self,
        seconds: float,
        fn: Callable,
        *args,
        loop: asyncio.AbstractEventLoop,
        name: str,
    ) -> None:
        self._seconds = seconds
        self._name = name
        self._event_loop = loop
        self._fn = fn
        self._args = args
        self._task: asyncio.Task | None = None

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self._seconds}, {self._fn.__name__})"

    @property
    def seconds(self) -> float:
        return self._seconds

    async def create(self) -> None:
        if self._task is not None:
            return
        logging.debug(f"Creating {self._name} task")
        self._task = self._event_loop.create_task(self._run())

    async def cancel(self) -> None:
        if self._task is None:
            return
        if not self._task.done():
            logging.debug(f"Canceling {self._name} task")
            self._task.cancel()
        try:
            result = await self._task
            logging.debug(f"Task {self._name} returned {result}")
        except asyncio.CancelledError:
            logging.debug(f"Task {self._name} has been cancelled")
        self._task = None

    async def _run(self) -> None:
        logging.debug(f"Start of {self._name} task")
        try:
            await asyncio.sleep(self._seconds)
        except asyncio.CancelledError:
            logging.debug(f"Wait in {self._name} task has been cancelled")
            return
        logging.debug(f"Triggering action in {self._name} task")
        result = self._fn(*self._args)
        if inspect.iscoroutinefunction(self._fn):
            await result
        logging.debug(f"End of {self._name} task")

    async def restart(self) -> None:
        await self.cancel()
        await self.create()


class GpioSample(BaseModel, extra="forbid"):
    state: int
    timestamp: int


class ClientInfo(BaseModel):
    id: str
    # TODO: implement configurable name
    status: ClientStatus | None = None
    # TODO: implement configurable pin invert ?
    monitored_gpio: tuple[str, ...] | None = None
    heartbeat: int | None = None
    heartbeat_watchdog: DelayedAction | None = None
    hardware: str | None = None
    buffer_total_dropped_item: int = 0
    gpio: dict[str, GpioSample] = Field(default_factory=dict)

    class Config:
        arbitrary_types_allowed = True



import asyncio
import errno
import logging
import socket
import ssl
from argparse import Namespace
from functools import partial
from typing import Any

import backoff
from paho.mqtt.client import (
    Client,
    ConnectFlags,
    DisconnectFlags,
    MQTTMessage,
    MQTTv311,
    MQTTMessageInfo,
    MQTT_LOG_INFO,
    MQTT_LOG_NOTICE,
    MQTT_LOG_WARNING,
    MQTT_LOG_ERR,
    MQTT_LOG_DEBUG,
    MQTT_ERR_NO_CONN,
    MQTT_ERR_SUCCESS,
    PayloadType,
)
from paho.mqtt.enums import CallbackAPIVersion
from paho.mqtt.properties import Properties
from paho.mqtt.reasoncodes import ReasonCodes

from gcn_manager import (
    AppError,
    MqttPublisher,
    MqttMessage,
    MessageProcessor,
    DelayedAction,
)
from gcn_manager.constants import *
from gcn_manager.notifiers import (
    MqttDisconnectedNotification,
    MqttStillConnectingNotification,
    MqttConnectedNotification,
)


class MqttMessageIdTracker:
    def __init__(self) -> None:
        self._topics: dict[int, str] = dict()

    def track(self, mid: int, topic: str) -> None:
        logging.debug(f"Tracking {mid=} {topic=}")
        if mid in self._topics:
            raise AppError(f"Wants to track mid {mid} but already tracked")
        self._topics[mid] = topic

    def untrack(self, mid: int) -> str:
        try:
            topic = self._topics.pop(mid)
        except KeyError:
            raise AppError(f"Wants to untrack mid {mid} but not tracked")
        logging.debug(f"Untracking {mid=} {topic=}")
        return topic


class MqttAgent(MqttPublisher):
    TASK_MESSAGE_ATTRIBUTE_PREFIX = "random_prefix_846731984_task_mqtt_message"

    def __init__(
        self,
        args: Namespace,
        client_id: str,
        *,
        processor: MessageProcessor,
        shutdown_requested: asyncio.Event,
        event_loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__(client_id)
        # dependencies
        self._args = args
        self._processor = processor
        self._shutdown_requested = shutdown_requested
        self._event_loop = event_loop
        # business logic
        self._remaining_tasks: set[asyncio.Task] = set()
        self._subscribed_topics: set[str] = set()
        # watchdog
        notification = MqttStillConnectingNotification(
            id=self.client_id,
            server=self._mqtt_server_str(),
            elapsed_seconds=self._args.mqtt_still_connecting_alert,
        )
        self._still_connecting_watchdog_delayed_action = DelayedAction(
            self._args.mqtt_still_connecting_alert,
            notification.send,
            loop=self._event_loop,
            name=f"{self.client_id}-connecting-watchdog",
        )
        # agent lifecycle
        self._mid = MqttMessageIdTracker()
        self._no_writer_left: asyncio.Event | None = None
        self._misc_loop_task: asyncio.Task | None = None
        self._connect_result: asyncio.Future | None = None
        self._disconnect_result: asyncio.Future | None = None
        # PAHO client configuration
        self._paho_mqtt_client = Client(
            callback_api_version=CallbackAPIVersion.VERSION2,
            clean_session=True,
            client_id=self.client_id,
            protocol=MQTTv311,
            transport=args.mqtt_transport,
            reconnect_on_failure=self._args.mqtt_reconnect,
        )
        context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
        if self._args.mqtt_tls_min_version is not None:
            context.minimum_version = self._args.mqtt_tls_min_version
        if self._args.mqtt_tls_max_version is not None:
            context.maximum_version = self._args.mqtt_tls_max_version
        if self._args.mqtt_tls_ciphers is not None:
            try:
                context.set_ciphers(args.mqtt_tls_ciphers)
            except ssl.SSLError as e:
                raise AppError(
                    f"Could not set MQTT TLS ciphers string {args.mqtt_tls_ciphers} : {e}"
                )
        self._paho_mqtt_client.tls_set_context(context)
        self._paho_mqtt_client.connect_timeout = self._args.mqtt_connect_timeout
        self._paho_mqtt_client.username_pw_set(
            username=self._args.mqtt_user_name, password=self._args.mqtt_user_password
        )
        self._paho_mqtt_client.on_connect = self._on_connect
        self._paho_mqtt_client.on_connect_fail = self._on_connect_fail
        self._paho_mqtt_client.on_disconnect = self._on_disconnect
        self._paho_mqtt_client.on_log = self._on_log
        self._paho_mqtt_client.on_message = self._on_message
        self._paho_mqtt_client.on_pre_connect = self._on_pre_connect
        self._paho_mqtt_client.on_publish = self._on_publish
        self._paho_mqtt_client.on_socket_close = self._on_socket_close
        self._paho_mqtt_client.on_socket_open = self._on_socket_open
        self._paho_mqtt_client.on_socket_register_write = self._on_socket_register_write
        self._paho_mqtt_client.on_socket_unregister_write = (
            self._on_socket_unregister_write
        )
        self._paho_mqtt_client.on_subscribe = self._on_subscribe
        self._paho_mqtt_client.on_unsubscribe = self._on_unsubscribe
        self._paho_mqtt_client.will_set(
            topic=self._get_manager_status_topic(),
            payload=MQTT_APP_MANAGER_STATUS_OFFLINE,
            qos=1,
            retain=True,
        )

    def _get_manager_status_topic(self) -> str:
        return f"{MQTT_APP_MANAGER_STATUS_TOPIC}/{self.client_id}"

    def _mqtt_server_str(self) -> str:
        return f"{self._args.mqtt_host}:{self._args.mqtt_port}"

    @backoff.on_exception(
        partial(backoff.expo, base=1.5, max_value=3),
        jitter=backoff.random_jitter,
        exception=(socket.gaierror, TimeoutError, ConnectionRefusedError, OSError),
    )
    def _connect(self) -> asyncio.Future:
        # interrupt the connection attempts when a shutdown has been requested
        if self._shutdown_requested.is_set():
            raise AppError("Still trying to connect while shutting down")

        # initialize the futures representing the connection and disconnection reasons
        logging.info(f"Connecting to MQTT broker '{self._mqtt_server_str()}'...")
        self._connect_result = self._event_loop.create_future()
        self._disconnect_result = self._event_loop.create_future()

        # Attempt to connect to the transport (only !) with back-off on retryable errors.
        # The following methods are implicitly called on the (possibly different from main) thread, then return
        # - _on_pre_connect
        # - _on_socket_open --> add socket to event loop readers + safely create loop task on loop
        # - _on_log
        # - _on_socket_register_write --> add socket to event loop writers (to send CONNECT message)
        try:
            self._paho_mqtt_client.connect(
                self._args.mqtt_host, self._args.mqtt_port, self._args.mqtt_keep_alive
            )
        except ssl.SSLError as e:
            # TLS errors are red flag for bad configuration or misbehaving actors : do not retry
            exc = AppError(
                f"TLS error while communicating with {self._args.mqtt_port} of '{self._args.mqtt_host}' : {e}"
            )
            self._connect_result.set_exception(exc)
            self._disconnect_result.set_exception(exc)
        except OSError as e:
            # only retry network unreachable errors
            if e.errno == errno.ENETUNREACH:
                raise
            raise AppError(
                f"OS error while trying to connect to MQTT broker '{self._args.mqtt_host}' : {e}"
            )

        # and return result if it did not fail temporarily
        return self._connect_result

    def publish(
        self,
        topic: str,
        payload: PayloadType,
        qos: int = 0,
        retain: bool = False,
        properties: Properties | None = None,
    ) -> MQTTMessageInfo:
        info = self._paho_mqtt_client.publish(topic, payload, qos, retain, properties)
        logging.debug(f"Published message {info}")
        self._mid.track(info.mid, topic)  # for _on_publish()
        return info

    def subscribe(self, topic: str, qos: int) -> None:
        error, mid = self._paho_mqtt_client.subscribe(topic, qos)
        if error == MQTT_ERR_NO_CONN:
            raise AppError(f"Could not subscribe to topic {topic}: not connected")
        elif error == MQTT_ERR_SUCCESS:
            logging.debug(f"Subscribing to topic {topic} {qos=}")
            self._mid.track(mid, topic)  # for _on_subscribe()
            self._subscribed_topics.add(topic)
        else:
            raise AppError(
                f"Unknown MQTTErrorCode during subscribe to {topic}: {error}"
            )

    def unsubscribe(
        self, topic: str | list[str], properties: Properties | None = None
    ) -> None:
        if topic not in self._subscribed_topics:
            raise AppError(
                f"Topic {topic} was not subscribed and cannot be unsubscribed"
            )
        error, mid = self._paho_mqtt_client.unsubscribe(topic, properties)
        if error == MQTT_ERR_NO_CONN:
            raise AppError(f"Could not unsubscribe to topic {topic}: not connected")
        elif error == MQTT_ERR_SUCCESS:
            logging.debug(f"Unsubscribing from topic {topic}")
            self._mid.track(mid, topic)  # for _on_unsubscribe()
            self._subscribed_topics.remove(topic)
        else:
            raise AppError(
                f"Unknown MQTTErrorCode during subscribe to {topic}: {error}"
            )

    def unsubscribe_all(self) -> None:
        topics = self._subscribed_topics.copy()  # build a shallow copy to iterate over
        logging.info(f"Unsubscribing to {len(topics)} subscribed MQTT topics")
        for topic in topics:
            # as this modifies self._subscribed_topics in-place
            self.unsubscribe(topic)

    def _disconnect(self) -> None:
        logging.debug("Disconnecting from broker")
        self._paho_mqtt_client.disconnect()

    def _set_status(self, *, online: bool) -> MQTTMessageInfo:
        topic = self._get_manager_status_topic()
        message = (
            MQTT_APP_MANAGER_STATUS_ONLINE
            if online
            else MQTT_APP_MANAGER_STATUS_OFFLINE
        )
        logging.debug(f"Setting {self.client_id} status to {message}")
        return self.publish(topic=topic, payload=message, qos=1, retain=True)

    @property
    def connection_result(self) -> asyncio.Future:
        return self._connect_result

    def _on_connect(
        self,
        _client: Client,
        _user_data: Any,
        connect_flags: ConnectFlags,
        reason_code: ReasonCodes,
        properties: Properties,
    ) -> None:
        reason_text = reason_code.getName()
        logging.debug(
            f"Connected, "
            f"reason code '{reason_code.value}/{reason_text}', "
            f"session_present {connect_flags.session_present}, "
            f"properties {properties}"
        )
        # resolve the future early, so that the reason_code is pertinent
        if reason_code.is_failure:
            exc = AppError(
                f"Failure upon connecting to '{self._args.mqtt_host}:{self._args.mqtt_port}': {reason_text}"
            )
            self._connect_result.set_exception(exc)
            return
        # memorize and communicate the result of the connection attempt
        self._connect_result.set_result(reason_code)

    def _on_connect_fail(self, _client: Client, _user_data: Any) -> None:
        """
        As per client library source code
        - this function is only called from client._handle_on_connect_fail()
        - _handle_on_connect_fail is only called from client.loop_forever()
        So when running in an async loop, this function will never not be called
        """
        raise AppError(
            "MqttClient._on_connect_fail has ben called : "
            "this should never happen if NOT using loop_forever() !"
        )

    def _on_disconnect(
        self,
        _client: Client,
        _user_data: Any,
        disconnect_flags: DisconnectFlags,
        reason_code: ReasonCodes,
        properties: Properties,
    ) -> None:
        logging.debug(
            f"Disconnected, "
            f"reason code '{reason_code.value}/{reason_code.getName()}', "
            f"session_present {disconnect_flags.is_disconnect_packet_from_server}, "
            f"properties {properties}"
        )
        # resolve the Future only if it has not already been resolved, for example, in on_connect on error
        if not self._disconnect_result.done():
            self._disconnect_result.set_result(reason_code)

    @staticmethod
    def _on_log(_client: Client, _user_data: Any, level: int, buf: str) -> None:
        if level == MQTT_LOG_INFO:
            level = "INFO"
        elif level == MQTT_LOG_NOTICE:
            level = "NOTICE"
        elif level == MQTT_LOG_WARNING:
            level = "WARNING"
        elif level == MQTT_LOG_ERR:
            level = "ERROR"
        elif level == MQTT_LOG_DEBUG:
            level = "DEBUG"
        else:
            raise AppError(f"Unknown MQTT log level {level}")
        logging.debug(f"MQTT log {level} {buf}")

    def _on_message(self, _client: Client, _user_data: Any, msg: MQTTMessage) -> None:
        logging.debug(
            f"Received message: "
            f"timestamp {msg.timestamp}, "
            f"state {msg.state}, "
            f"dup {msg.dup}, "
            f"mid {msg.mid}, "
            f"qos {msg.qos}, "
            f"retain {msg.retain}, "
            # used with MQTTMessageInfo.wait_for_publish(timeout_sec)
            f"info {msg.info}, "
            f"properties {msg.properties}, "
            f"topic {msg.topic}, "
            f"payload {msg.payload}"
        )
        # noinspection PyTypeChecker
        message = MqttMessage(topic=msg.topic, payload=msg.payload)
        future = self._processor.process(message, publisher=self)
        # create a task per message
        task = self._event_loop.create_task(future)
        # patching the message onto the task object. Possibly use a more random prefix to prevent name conflict ?
        setattr(task, self.TASK_MESSAGE_ATTRIBUTE_PREFIX, message)
        # store a strong reference to the created task, to prevent garbage collection in asyncio
        self._remaining_tasks.add(task)

    @staticmethod
    def _on_pre_connect(_client: Client, _user_data: Any) -> None:
        logging.debug(f"Starting connection.")

    def _on_publish(
        self,
        _client: Client,
        _user_data: Any,
        mid: int,
        reason_code: ReasonCodes,
        properties: Properties,
    ) -> None:
        logging.debug(
            f"Broker responded to publish {mid=} {reason_code=} {str(properties)=}"
        )
        topic = self._mid.untrack(mid)  # from publish()
        # FIXME: see https://github.com/eclipse-paho/paho.mqtt.python/issues/895
        if reason_code.is_failure:
            raise AppError(
                f"Encountered {reason_code.value}/{reason_code} while publishing to {topic}"
            )

    def _on_socket_close(self, _client: Client, _user_data: Any, sock) -> None:
        """
        cancel task for asynchronous reading of packets, no more incoming messages can arrive
        cancel task for queuing keepalive packets
        """
        logging.debug(
            f"Socket {sock} is about to close, remove socket from loop readers"
        )
        self._event_loop.remove_reader(sock)
        if self._misc_loop_task is not None:
            self._misc_loop_task.cancel()

    def _loop_read(self) -> None:
        """task for asynchronous reading of incoming socket data, then calling the appropriate callbacks"""
        logging.debug("Socket is readable, calling loop_read")
        self._paho_mqtt_client.loop_read()

    def _on_socket_open(self, _client: Client, _user_data: Any, sock) -> None:
        """
        set socket write buffer size
        add task for asynchronous reading of packets for incoming messages
        add task for queuing keepalive packets as needed
        """
        logging.debug(
            f"Socket {sock} opened, add socket to loop readers, set sock opt, create misc loop async task"
        )
        # noinspection PyUnresolvedReferences
        _set_sock_opt = self._paho_mqtt_client.socket().setsockopt
        _set_sock_opt(
            socket.SOL_SOCKET, socket.SO_SNDBUF, self._args.mqtt_socket_send_buffer_size
        )
        # noinspection PyTypeChecker
        self._event_loop.add_reader(sock, self._loop_read)
        self._misc_loop_task = asyncio.run_coroutine_threadsafe(
            self._loop_misc(), self._event_loop
        )

    def _loop_write(self) -> None:
        """task for asynchronous writing of buffered packets to socket, as long as there are any to send"""
        logging.debug("Socket is writable, calling loop_write")
        self._paho_mqtt_client.loop_write()

    def _on_socket_register_write(self, _client: Client, _user_data: Any, sock) -> None:
        logging.debug(f"Client data needs writing to {sock}, add sock to loop writers")
        # noinspection PyTypeChecker
        self._event_loop.add_writer(sock, self._loop_write)
        self._no_writer_left = asyncio.Event()

    def _on_socket_unregister_write(
        self, _client: Client, _user_data: Any, sock
    ) -> None:
        logging.debug(
            f"No more client data to write into socket {sock}, remove sock from loop writers"
        )
        self._event_loop.remove_writer(sock)
        self._no_writer_left.set()

    def _on_subscribe(
        self,
        _client: Client,
        _user_data: Any,
        mid: int,
        reason_code_list: list[ReasonCodes],
        properties: Properties,
    ) -> None:
        logging.debug(
            f"Broker responded to subscribe {mid=} {reason_code_list=} {str(properties)=}"
        )
        topic = self._mid.untrack(mid)  # from subscribe()
        for reason_code in reason_code_list:
            if reason_code.is_failure:
                raise AppError(
                    f"Encountered {reason_code.value}/{reason_code} while subscribing to {topic}"
                )

    def _on_unsubscribe(
        self,
        _client: Client,
        _user_data: Any,
        mid: int,
        reason_code_list: list[ReasonCodes],
        properties: Properties,
    ) -> None:
        logging.debug(
            f"Broker responded to unsubscribe {mid=} {reason_code_list=} {str(properties)=}"
        )
        topic = self._mid.untrack(mid)  # from unsubscribe()
        for reason_code in reason_code_list:
            if reason_code.is_failure:
                raise AppError(
                    f"Encountered {reason_code.value}/{reason_code} while unsubscribing to {topic}"
                )

    async def _loop_misc(self) -> None:
        """asynchronous task for queuing keepalive packets as needed regarding message activity"""
        logging.debug("MQTT misc_loop started")
        while self._paho_mqtt_client.loop_misc() == MQTT_ERR_SUCCESS:
            try:
                # nothing fast is required, only for keepalive
                await asyncio.sleep(self._args.idle_loop_sleep)
            except asyncio.CancelledError:
                logging.debug("MQTT misc_loop cancelled")
                break
        logging.debug("MQTT misc_loop finished")

    async def _cleanup_finished_handling_tasks(self, *, timeout: float | None) -> None:
        if (
            len(self._remaining_tasks) == 0
        ):  # asyncio.wait() raises ValueError on empty set
            return
        finished_tasks, self._remaining_tasks = await asyncio.wait(
            self._remaining_tasks, timeout=timeout, return_when=asyncio.FIRST_COMPLETED
        )
        if len(finished_tasks) > 0:
            logging.debug(
                f"Found {len(self._remaining_tasks)} remaining tasks and {len(finished_tasks)} finished"
            )

        while finished_tasks:
            task = finished_tasks.pop()
            exc = task.exception()
            if exc is None:
                continue
            message = getattr(task, self.TASK_MESSAGE_ATTRIBUTE_PREFIX, None)
            logging.warning(f"Ignoring message {message} due to exception : {exc}")

    async def run_once(self) -> None:
        # start the watchdog task which will alert if the connection attempt does not go through in due time
        await self._still_connecting_watchdog_delayed_action.create()

        # try to connect
        try:
            connect_result_future = await asyncio.to_thread(self._connect)
            reason_code = await connect_result_future
        except asyncio.CancelledError:
            logging.warning(
                "MQTT Agent app task was cancelled while waiting for connection"
            )
            raise
        except AppError as e:
            logging.error(f"Connection failed irrevocably : {e}")
            raise
        finally:
            # cancel the warning task
            await self._still_connecting_watchdog_delayed_action.cancel()

        # connection established successfully, publish our presence and subscribe to desired topics
        logging.info(f"Connected to MQTT broker : {reason_code.value}/{reason_code}")
        await MqttConnectedNotification(
            id=self.client_id, server=self._mqtt_server_str()
        ).send()
        self._set_status(online=True)
        self.subscribe(MQTT_APP_MANAGER_STATUS_SUBSCRIPTION, 1)
        self.subscribe(f"{MQTT_APP_CLIENT_SUBSCRIPTION}/+", 1)
        self.subscribe(f"{MQTT_APP_CLIENT_SUBSCRIPTION}/{MQTT_APP_CLIENT_GPIO}/+", 1)

        # periodically cleanup finished tasks until shutdown or spurious disconnection
        while not self._shutdown_requested.is_set():
            if not self._paho_mqtt_client.is_connected():
                logging.warning("Detected unexpected disconnection.")
                await MqttDisconnectedNotification(
                    id=self.client_id, server=self._mqtt_server_str()
                ).send()
                break
            try:
                await self._cleanup_finished_handling_tasks(
                    timeout=self._args.idle_loop_sleep
                )
                # nothing fast is required, cleanup only
                await asyncio.sleep(self._args.idle_loop_sleep)
            except asyncio.CancelledError:
                logging.error(
                    "MQTT Agent app task was cancelled before shutdown was requested"
                )
                return

        # cleanup subscription and status if still connected
        if self._paho_mqtt_client.is_connected():
            logging.info(f"MQTT task detected shutdown request")
            # unsubscribing to all tracked topics to quench the flow
            self.unsubscribe_all()
            # update our online status
            self._set_status(online=False)
            # waiting for writer tasks to finish
            await self._no_writer_left.wait()
            logging.info(f"No more writers, disconnecting from MQTT server")

        # connected or not, wait for all tasks to finish
        while self._remaining_tasks:
            logging.info(
                f"Waiting for {len(self._remaining_tasks)} handling tasks to finish..."
            )
            try:
                # minimum display during cleanup
                await self._cleanup_finished_handling_tasks(timeout=None)
            except asyncio.CancelledError:
                logging.error(
                    "MQTT Agent app task was cancelled before handling tasks could finish"
                )
                return

        if self._paho_mqtt_client.is_connected():
            # disconnect cleanly
            self._disconnect()
            try:
                reason_code = await self._disconnect_result
                logging.info(
                    f"Disconnected from MQTT broker {reason_code.value}/{reason_code}"
                )
            except asyncio.CancelledError:
                logging.debug(
                    "MQTT Agent app task was cancelled while waiting for disconnection"
                )
                return
            except AppError as e:
                logging.error(f"Disconnection failed irrevocably : {e}")
                return
        logging.info("MQTT lifecycle finished")
